from functools import partial
from typing import Callable, Union
from typing import List

from stratosphere.store.db.database import next_ulid
from stratosphere.utils.time import current_time


class Run:
    """A run represents an instance of the experiment, obtained by
    combining the fixed and variable parameters. The Run objects
    (and the tracked attributes)  must be serializable with cloudpickle.
    """

    def __init__(
        self,
        kwargs: dict = None,
        parameters: dict = None,
        attributes: dict = None,
        blobs: dict = None,
        funcs: Union[Callable, List[Callable]] = None,
        experiment_name: str = None,
    ):
        """Create a new run.

        Args:
            kwargs (dict, optional): Fixed parameters for all runs of an experiment. Defaults to None.
            parameters (dict, optional): Variable parameters to be considered for this run. Defaults to None.
            attributes (dict, optional): Attributes to be tracked. Defaults to None.
            blobs (dict, optional): Binary attributes to be tracked. Defaults to None.
            funcs (Union[Callable, List[Callable]], optional): One or more functions to be executed. Their only parameter is an instance of
            the Run class itself.
            experiment_name (str, optional): Name of the experiment the run belongs to. Defaults to None.
        """

        self.kwargs = {} if kwargs is None else kwargs
        self.parameters = {} if parameters is None else parameters
        self.attributes = {} if attributes is None else attributes
        self.blobs = {} if blobs is None else blobs
        self.funcs = [] if funcs is None else funcs
        self.execution_time = None
        self.completion_time = None
        self.run_id = next_ulid()
        self.experiment_name = experiment_name

    @classmethod
    def execute(cls, run):
        """Execute the functions on the Run object. The functions can modify its .attributes and .blobs
        fields, that are treated as state accumulators. Other changes to the Run object might be lost
        if pickling and moving the Run, therefore they are discouraged.

        Args:
            run (Run): Run we want to execute.

        Returns:
            Run: Run object after the execution of the functions.
        """

        run.execution_time = current_time()

        # The funcs field might be one or more functions.
        if callable(run.funcs):
            run.funcs(run)
        else:
            [func(run) for func in run.funcs]

        run.completion_time = current_time()
        return run

    @classmethod
    def execute_func(
        cls,
        run,
        override_funcs: Union[Callable, List[Callable]] = None,
    ):
        """_summary_

        Args:
            run (_type_): _description_
            override_funcs (Union[Callable, List[Callable]], optional):
            Function(s) to be executed, overriding the already configured ones.
            Important, the new functions are overwriting the existing ones.


        Returns:
            Callable: function ready to be called, to execute the run.
        """
        if override_funcs:
            run.funcs = override_funcs
        return partial(cls.execute, run)



import logging
import uuid
from typing import Callable, Iterator, List, Union

import pandas as pd
from sqlalchemy.orm import sessionmaker, Query
from sqlalchemy import create_engine, MetaData, Table, sql
from sqlalchemy.engine import make_url
from sqlalchemy.exc import NoSuchTableError
from stratosphere.utils.progress import progress
from ulid import monotonic as ulid

from stratosphere import config
from stratosphere.utils.enums import IfExists
from stratosphere.store.db.schema import Base


class Database:
    """
    This class is the entry point for all things SQLAlchemy, and it represents the link
    to the database for experiments and runs. It uses the SQAlchemy 1.4 API interface.
    """

    def __init__(self, url: str = config.db.default_url, echo=False, lazy=False):
        """It creates a new database interface.

        Args:
            url (str, optional): Database URL, passed to SQLAlchemy. Defaults to config.db.default_url.
            echo (bool, optional): If True, log the SQL commands. Defaults to False.
            lazy (bool, optional): If True, don't initialize the object (useful to handle serialization).
            Defaults to False.
        """

        if lazy:
            # We activate this flag in the __reduce__, called upon pickling.
            # This behaviour allows to use sqlite in memory, since only one sqlite instance
            # gets created and reused for all experiments, runs, created or loaded via unpickling.
            # Further, it let us not run into pickling errors, which do happen with SQLAlchemy objects.
            return

        self.echo = echo
        self.url = make_url(url)
        self.info()

        # Set up database connector, without establishing any connection yet
        # https://docs.sqlalchemy.org/en/13/core/pooling.html#disconnect-handling-pessimistic
        self.engine = create_engine(url, echo=echo, pool_pre_ping=True)

        # Session factory
        self.session = sessionmaker(self.engine)

        # create tables if missing
        Base.metadata.create_all(self.engine)

    def __reduce__(self):
        """
        We return a tuple of class_name to call,
        and optional parameters to pass when re-creating
        """
        return self.__class__, (self.url, self.echo, True)

    def info(self):
        """Log some stats about the database."""
        logging.info(f"Database: {self.url.render_as_string(hide_password=True)}")

    def delete(self):
        """Delete all tables defined in the schema module"""
        with self.session() as session:
            for table in Base.metadata.sorted_tables:
                session.query(table).delete()
            session.commit()

    def pandas_to_sql(self, df: pd.DataFrame, name: str, if_exists: IfExists):
        """Insert a Pandas dataframe as a new database table.

        Args:
            df (pd.DataFrame): Dataframe to insert.
            name (str): Name of the database table to create.
            if_exists (IfExists): Either "replace" or "fail".
        """

        if len(df) == 0:
            # nothing to do
            return

        with self.session() as session:
            if config.tqdm.enable:
                dfs = chunker(df, config.db.query_write_chunk_size)
                funcs = []
                for idx, df_chunk in enumerate(dfs):

                    def process_chunk():
                        df_chunk.to_sql(
                            name,
                            session.bind,
                            if_exists=if_exists if idx == 0 else "append",
                            index=False,
                        )
                        return len(df_chunk), None

                    funcs.append(process_chunk)

                tqdm_chunk(funcs, len(df))
            else:
                df.to_sql(
                    f"experiment.{name}", session.bind, if_exists=if_exists, index=False
                )

    def drop_table(self, name: str):
        """Drop database table if it exists.

        Args:
            name (str): Name of the table to drop.
        """
        with self.session() as session:
            try:
                Table(
                    name, MetaData(bind=session.bind), autoload_with=session.bind
                ).drop()
            except NoSuchTableError:
                pass
            session.commit()

    def vacuum(self):
        """Vacuum the database. This operation makes the database
        more compact and performant.
        """
        with self.session() as session:
            if self.url.drivername not in ["sqlite"]:
                # sqlite wants a transaction to VACUUM,
                # postgresql doesn't. With a COMMIT,
                # we terminate the transaction, and
                # we can then execute the VACUUM command.
                session.execute("COMMIT")

            session.execute("VACUUM")

        logging.info("VACUUM executed.")

    def pandas(
        self,
        query: Union[str, Query, sql.expression.text],
        verbose=False,
        use_tqdm=True,
        tqdm_total=None,
    ) -> pd.DataFrame:
        """Query database and return result as Pandas dataframe.

        Args:
            query (Union[str, Query, sql.expression.text]): Query to execute.
            verbose (bool, optional): If True, log the SQL query. Defaults to False.
            use_tqdm (bool, optional): If True, use tqdm progress bar. Defaults to True.
            tqdm_total (_type_, optional): Provide the expected number of rows retrieved.
            If missing, we'll first count them, and then retrieve them, resulting in two
            distinct queries. Defaults to None.

        Returns:
            pd.DataFrame: Resulting rows.
        """

        with self.session() as session:
            return pandas_query(
                query,
                session,
                verbose=verbose,
                use_tqdm=use_tqdm,
                tqdm_total=tqdm_total,
            )


def pandas_query(
    query: Union[str, Query, sql.expression.text],
    session,
    verbose=False,
    use_tqdm=True,
    tqdm_total=None,
) -> pd.DataFrame:
    """Evaluate an SQL query on the database and return the result
    as a Pandas dataframe.

    Args:
        query (Union[str, Query, sql.expression.text]): Query to execute.
        session: SQLAlchemy session to use.
        verbose (bool, optional): If True, log the SQL query. Defaults to False.
        use_tqdm (bool, optional): If True, use tqdm progress bar. Defaults to True.
        tqdm_total (_type_, optional): Provide the expected number of rows retrieved.
        If missing, we'll first count them, and then retrieve them, resulting in two
        distinct queries. Defaults to None.

    Returns:
        pd.DataFrame: Resulting rows. The column types are guessed by Pandas.
    """

    # Normalise query to sql.expression.text
    if type(query) == Query:
        query = query.statement
    elif type(query) == str:
        query = sql.expression.text(query)

    if verbose:
        logging.info(f"SQL: {query.compile(session.bind)}")

    if use_tqdm:
        if tqdm_total is None:
            # If hint on total not available, query the database for it
            tqdm_total = session.execute(f"SELECT COUNT(*) FROM ({query}) t").first()[0]

        df_chunks = pd.read_sql(
            query, session.bind, chunksize=config.db.query_read_chunk_size
        )

        funcs = [lambda: (len(df_chunk), df_chunk) for df_chunk in df_chunks]
        dfs = tqdm_chunk(funcs, tqdm_total)

        df = pd.concat(dfs, ignore_index=True)
    else:
        df = pd.read_sql(query, session.bind)

    return df.apply(pd.to_numeric, errors="ignore")


def next_ulid() -> str:
    """Return the next monotonic ULID (https://github.com/ulid/spec).

    Returns:
        str: New ULID in UUID format, ready to be used as Experiment or Run ID.
    """
    return str(uuid.UUID(bytes=ulid.new().bytes))


def chunker(seq: List, size) -> List[List]:
    """Given a list, return a list of lists, where each sub-list is up to a certain
    chunk size.

    Args:
        seq (List): List to partition.
        size (_type_): Chunk size.

    Returns:
        List[List]: List of chunks.
    """
    return (seq[pos : pos + size] for pos in range(0, len(seq), size))


def tqdm_chunk(
    iterator: Iterator[Callable],
    total: int,
) -> List:
    """Renders a progress bar, working on chunks of work.

    Args:
        iterator (Iterator[Callable]): Iterator of functions to call, returning (size, ret). Size is
        used to update the progress bar, and ret is accumulated as one more element in the returned list.
        total (int): Total number of items to process, regardless of chunking.

    Returns:
        List: List of return values from callables.
    """

    pbar = progress(total=total)
    pbar.clear()

    rets = []
    for item in iterator:
        size, ret = item()
        pbar.update(size)

        rets.append(ret)

    pbar.close()

    return rets



from enum import Enum

import sqlalchemy as sa
from sqlalchemy.orm import declarative_base
from sqlalchemy_utils.types.uuid import UUIDType

# Construct a base class for declarative class definitions.
Base = declarative_base()

# Enum defining the status of an experiment.
ExperimentStatus = Enum(
    "ExperimentStatus", ["UNKNOWN", "COMPLETED", "FAILED", "INTERRUPTED"]
)

# UUID type, to be used in model definitions. By passing binary=False,
# we fall back to the string representation of UUIDs if there's no native type
# (as in SQLite).
uuid_type = UUIDType(binary=False)


class Experiment(Base):
    """Model representing an experiment in the database."""

    __tablename__ = "experiments"
    id = sa.Column(uuid_type, primary_key=True, default=None)
    name = sa.Column(sa.String(250), nullable=False, unique=True)
    execution_time = sa.Column(sa.BigInteger(), nullable=False, default=None)
    completion_time = sa.Column(sa.BigInteger(), nullable=False, default=None)
    attributes = sa.Column(sa.String(), nullable=False, default=None)
    serialized = sa.Column(sa.LargeBinary(), nullable=True, default=None)
    # Currently, the status is not used, and it's always set to UNKNOWN.
    # TODO: either use it, or drop it.
    status = sa.Column(
        sa.Enum(ExperimentStatus),
        nullable=False,
        default=ExperimentStatus["UNKNOWN"],
    )



import json
import zlib

import cloudpickle


def pickle_dumps(obj: object) -> bytes:
    """It returns the compressed (zlib) pickled object. The Pickle DEFAULT_PROTOCOL is used
    for maximum compatibility.

    Args:
        obj (object): Object to serialize.

    Returns:
        bytes: Compressed serialized object.
    """
    return zlib.compress(cloudpickle.dumps(obj, protocol=cloudpickle.DEFAULT_PROTOCOL))


def pickle_loads(data: bytes) -> object:
    """It returns the loaded object, afer uncompressing and unpickling it.

    Args:
        data (bytes): Serialized object.

    Returns:
        object: Unserialized object.
    """
    return cloudpickle.loads(zlib.decompress(data))


def pickle_size(obj: object, unit: str = "b") -> int:
    """It returns the size of the object once serialised (including compression).

    Args:
        obj (object): Object to analyse.
        unit (str, optional): Unit of measure, b (Bytes), kb (KiloBytes), mb (MegaBytes). Defaults to "b".

    Returns:
        int: Size of the serialized object.
    """
    size_object = len(pickle_dumps(obj))
    if unit == "b":
        return int(size_object * 1e2) / 1e2
    elif unit == "kb":
        return int(size_object * 1e-3 * 1e2) / 1e2
    elif unit == "mb":
        return int(size_object * 1e-6 * 1e2) / 1e2
    else:
        return None


def json_dumps(obj: object) -> str:
    """Serialize an object to its JSON string representation.

    Args:
        obj (object): Object to serialize.

    Returns:
        str: Serialized object.
    """
    return json.dumps(obj)


def json_loads(obj: str) -> object:
    """Unserialize an object from its JSON string representation.

    Args:
        obj (str): Object to unserialize.

    Returns:
        object: Unserialized object.
    """
    return json.loads(obj)



import warnings
import sys

try:
    import google.colab

    in_colab = True
except ModuleNotFoundError:
    in_colab = False


if "pyodide" in sys.modules:
    in_pyodide = True
else:
    in_pyodide = False

with warnings.catch_warnings():
    warnings.simplefilter("ignore")

    if in_colab:
        from tqdm import tqdm
    else:
        from tqdm.auto import tqdm

    if in_pyodide:
        tqdm.monitor_interval = 0


def progress(*args, **kwargs):
    return tqdm(*args, **kwargs)
